from setuptools import setup

setup(
    name='KlebLib',
    version='1.0.0',
    author='Caleb Robson',
    author_email='robson.caleb.299@gmail.com',
    packages=['KlebLib'],
    url='http://pypi.python.org/pypi/KlebLib/',
    license='MIT',
    description='A collection of classes and functions that no-one will ever need.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown'
)
